/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { BST, BSTNode } from './bst';
import type { AVLTreeNodeNested, AVLTreeOptions, BiTreeDeleteResult, BTNKey } from '../../types';
import { BTNCallback } from '../../types';
import { IBinaryTree } from '../../interfaces';
export declare class AVLTreeNode<V = any, N extends AVLTreeNode<V, N> = AVLTreeNodeNested<V>> extends BSTNode<V, N> {
    height: number;
    constructor(key: BTNKey, value?: V);
}
export declare class AVLTree<V = any, N extends AVLTreeNode<V, N> = AVLTreeNode<V, AVLTreeNodeNested<V>>> extends BST<V, N> implements IBinaryTree<V, N> {
    /**
     * This is a constructor function for an AVL tree data structure in TypeScript.
     * @param {AVLTreeOptions} [options] - The `options` parameter is an optional object that can be passed to the
     * constructor of the AVLTree class. It allows you to customize the behavior of the AVL tree by providing different
     * options.
     */
    constructor(options?: AVLTreeOptions);
    /**
     * The function creates a new AVL tree node with the specified key and value.
     * @param {BTNKey} key - The key parameter is the key value that will be associated with
     * the new node. It is used to determine the position of the node in the binary search tree.
     * @param [value] - The parameter `value` is an optional value that can be assigned to the node. It is of
     * type `V`, which means it can be any value that is assignable to the `value` property of the
     * node type `N`.
     * @returns a new AVLTreeNode object with the specified key and value.
     */
    createNode(key: BTNKey, value?: V): N;
    /**
     * The function overrides the add method of a binary tree node and balances the tree after inserting
     * a new node.
     * @param {BTNKey | N | undefined} keyOrNode - The `keyOrNode` parameter can accept either a
     * `BTNKey` or a `N` (which represents a node in the binary tree) or `null`.
     * @param [value] - The `value` parameter is the value that you want to assign to the new node that you
     * are adding to the binary search tree.
     * @returns The method is returning the inserted node (`N`), `null`, or `undefined`.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    /**
     * The function overrides the delete method of a binary tree and balances the tree after deleting a
     * node if necessary.
     * @param {ReturnType<C>} identifier - The `identifier` parameter is either a
     * `BTNKey` or a generic type `N`. It represents the property of the node that we are
     * searching for. It can be a specific key value or any other property of the node.
     * @param callback - The `callback` parameter is a function that takes a node as input and returns a
     * value. This value is compared with the `identifier` parameter to determine if the node should be
     * included in the result. The `callback` parameter has a default value of
     * `this._defaultOneParamCallback`
     * @returns The method is returning an array of `BiTreeDeleteResult<N>` objects.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback?: C): BiTreeDeleteResult<N>[];
    /**
     * The function swaps the key, value, and height properties between two nodes in a binary tree.
     * @param {N} srcNode - The `srcNode` parameter represents the source node that needs to be swapped
     * with the `destNode`.
     * @param {N} destNode - The `destNode` parameter represents the destination node where the values
     * from the source node (`srcNode`) will be swapped to.
     * @returns The method is returning the `destNode` after swapping its properties with the `srcNode`.
     */
    protected _swap(srcNode: BTNKey | N | undefined, destNode: BTNKey | N | undefined): N | undefined;
    /**
     * The function calculates the balance factor of a node in a binary tree.
     * @param {N} node - The parameter "node" represents a node in a binary tree data structure.
     * @returns the balance factor of a given node. The balance factor is calculated by subtracting the
     * height of the left subtree from the height of the right subtree.
     */
    protected _balanceFactor(node: N): number;
    /**
     * The function updates the height of a node in a binary tree based on the heights of its left and
     * right children.
     * @param {N} node - The parameter "node" represents a node in a binary tree data structure.
     */
    protected _updateHeight(node: N): void;
    /**
     * The `_balancePath` function is used to update the heights of nodes and perform rotation operations
     * to restore balance in an AVL tree after inserting a node.
     * @param {N} node - The `node` parameter in the `_balancePath` function represents the node in the
     * AVL tree that needs to be balanced.
     */
    protected _balancePath(node: N): void;
    /**
     * The function `_balanceLL` performs a left-left rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceLL(A: N): void;
    /**
     * The `_balanceLR` function performs a left-right rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceLR(A: N): void;
    /**
     * The function `_balanceRR` performs a right-right rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceRR(A: N): void;
    /**
     * The function `_balanceRL` performs a right-left rotation to balance a binary tree.
     * @param {N} A - A is a node in a binary tree.
     */
    protected _balanceRL(A: N): void;
}
