/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BinaryTreeNodeNested, BinaryTreeOptions, BTNCallback, BTNKey } from '../../types';
import { BiTreeDeleteResult, DFSOrderPattern, FamilyPosition, IterationType } from '../../types';
import { IBinaryTree } from '../../interfaces';
/**
 * Represents a node in a binary tree.
 * @template V - The type of data stored in the node.
 * @template N - The type of the family relationship in the binary tree.
 */
export declare class BinaryTreeNode<V = any, N extends BinaryTreeNode<V, N> = BinaryTreeNode<V, BinaryTreeNodeNested<V>>> {
    /**
     * The key associated with the node.
     */
    key: BTNKey;
    /**
     * The value stored in the node.
     */
    value?: V;
    /**
     * The parent node of the current node.
     */
    parent?: N | null;
    /**
     * Creates a new instance of BinaryTreeNode.
     * @param {BTNKey} key - The key associated with the node.
     * @param {V} value - The value stored in the node.
     */
    constructor(key: BTNKey, value?: V);
    protected _left?: N | null;
    /**
     * Get the left child node.
     */
    get left(): N | null | undefined;
    /**
     * Set the left child node.
     * @param {N | null | undefined} v - The left child node.
     */
    set left(v: N | null | undefined);
    protected _right?: N | null;
    /**
     * Get the right child node.
     */
    get right(): N | null | undefined;
    /**
     * Set the right child node.
     * @param {N | null | undefined} v - The right child node.
     */
    set right(v: N | null | undefined);
    /**
     * Get the position of the node within its family.
     * @returns {FamilyPosition} - The family position of the node.
     */
    get familyPosition(): FamilyPosition;
}
/**
 * Represents a binary tree data structure.
 * @template N - The type of the binary tree's nodes.
 */
export declare class BinaryTree<V = any, N extends BinaryTreeNode<V, N> = BinaryTreeNode<V, BinaryTreeNodeNested<V>>> implements IBinaryTree<V, N> {
    iterationType: IterationType;
    /**
     * Creates a new instance of BinaryTree.
     * @param {BinaryTreeOptions} [options] - The options for the binary tree.
     */
    constructor(options?: BinaryTreeOptions);
    protected _root?: N | null;
    /**
     * Get the root node of the binary tree.
     */
    get root(): N | null | undefined;
    protected _size: number;
    /**
     * Get the number of nodes in the binary tree.
     */
    get size(): number;
    /**
     * Creates a new instance of BinaryTreeNode with the given key and value.
     * @param {BTNKey} key - The key for the new node.
     * @param {V} value - The value for the new node.
     * @returns {N} - The newly created BinaryTreeNode.
     */
    createNode(key: BTNKey, value?: V): N;
    /**
     * Add a node with the given key and value to the binary tree.
     * @param {BTNKey | N | null} keyOrNode - The key or node to add to the binary tree.
     * @param {V} value - The value for the new node (optional).
     * @returns {N | null | undefined} - The inserted node, or null if nothing was inserted, or undefined if the operation failed.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | null | undefined;
    /**
     * The `addMany` function takes an array of binary tree node IDs or nodes, and optionally an array of corresponding data
     * values, and adds them to the binary tree.
     * @param {(BTNKey | null)[] | (N | null)[]} keysOrNodes - An array of BTNKey or BinaryTreeNode
     * objects, or null values.
     * @param {V[]} [values] - The `values` parameter is an optional array of values (`V[]`) that corresponds to
     * the nodes or node IDs being added. It is used to set the value of each node being added. If `values` is not provided,
     * the value of the nodes will be `undefined`.
     * @returns The function `addMany` returns an array of `N`, `null`, or `undefined` values.
     */
    addMany(keysOrNodes: (BTNKey | N | null | undefined)[], values?: (V | undefined)[]): (N | null | undefined)[];
    /**
     * The `refill` function clears the binary tree and adds multiple nodes with the given IDs or nodes and optional data.
     * @param {(BTNKey | N)[]} keysOrNodes - The `keysOrNodes` parameter is an array that can contain either
     * `BTNKey` or `N` values.
     * @param {N[] | Array<V>} [values] - The `data` parameter is an optional array of values that will be assigned to
     * the nodes being added. If provided, the length of the `data` array should be equal to the length of the `keysOrNodes`
     * array. Each value in the `data` array will be assigned to the
     * @returns The method is returning a boolean value.
     */
    refill(keysOrNodes: (BTNKey | N | null | undefined)[], values?: (V | undefined)[]): boolean;
    delete<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C): BiTreeDeleteResult<N>[];
    delete<C extends BTNCallback<N, N>>(identifier: N | null | undefined, callback?: C): BiTreeDeleteResult<N>[];
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C): BiTreeDeleteResult<N>[];
    /**
     * The function `getDepth` calculates the depth of a given node in a binary tree relative to a
     * specified root node.
     * @param {BTNKey | N | null | undefined} distNode - The `distNode` parameter represents the node
     * whose depth we want to find in the binary tree. It can be either a node object (`N`), a key value
     * of the node (`BTNKey`), or `null`.
     * @param {BTNKey | N | null | undefined} beginRoot - The `beginRoot` parameter represents the
     * starting node from which we want to calculate the depth. It can be either a node object or the key
     * of a node in the binary tree. If no value is provided for `beginRoot`, it defaults to the root
     * node of the binary tree.
     * @returns the depth of the `distNode` relative to the `beginRoot`.
     */
    getDepth(distNode: BTNKey | N | null | undefined, beginRoot?: BTNKey | N | null | undefined): number;
    /**
     * The `getHeight` function calculates the maximum height of a binary tree using either recursive or
     * iterative approach.
     * @param {BTNKey | N | null | undefined} beginRoot - The `beginRoot` parameter represents the
     * starting node from which the height of the binary tree is calculated. It can be either a node
     * object (`N`), a key value of a node in the tree (`BTNKey`), or `null` if no starting
     * node is specified. If `
     * @param iterationType - The `iterationType` parameter is used to determine whether to calculate the
     * height of the binary tree using a recursive approach or an iterative approach. It can have two
     * possible values:
     * @returns the height of the binary tree.
     */
    getHeight(beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): number;
    /**
     * The `getMinHeight` function calculates the minimum height of a binary tree using either a
     * recursive or iterative approach.
     * @param {N | null | undefined} beginRoot - The `beginRoot` parameter is the starting node from which we want to
     * calculate the minimum height of the tree. It is optional and defaults to the root of the tree if
     * not provided.
     * @param iterationType - The `iterationType` parameter is used to determine the method of iteration
     * to calculate the minimum height of a binary tree. It can have two possible values:
     * @returns The function `getMinHeight` returns the minimum height of a binary tree.
     */
    getMinHeight(beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): number;
    /**
     * The function checks if a binary tree is perfectly balanced by comparing the minimum height and the
     * height of the tree.
     * @param {N | null | undefined} beginRoot - The parameter `beginRoot` is of type `N | null | undefined`, which means it can
     * either be of type `N` (representing a node in a tree) or `null` (representing an empty tree).
     * @returns The method is returning a boolean value.
     */
    isPerfectlyBalanced(beginRoot?: BTNKey | N | null | undefined): boolean;
    getNodes<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, onlyOne?: boolean, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N[];
    getNodes<C extends BTNCallback<N, N>>(identifier: N | null | undefined, callback?: C, onlyOne?: boolean, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N[];
    getNodes<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C, onlyOne?: boolean, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N[];
    has<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): boolean;
    has<C extends BTNCallback<N, N>>(identifier: N | null | undefined, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): boolean;
    has<C extends BTNCallback<N>>(identifier: ReturnType<C> | null | undefined, callback: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): boolean;
    getNode<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    getNode<C extends BTNCallback<N, N>>(identifier: N | null | undefined, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    getNode<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    /**
     * The function `getNodeByKey` searches for a node in a binary tree by its key, using either
     * recursive or iterative iteration.
     * @param {BTNKey} key - The `key` parameter is the key value that we are searching for in the tree.
     * It is used to find the node with the matching key value.
     * @param iterationType - The `iterationType` parameter is used to determine whether the search for
     * the node with the given key should be performed iteratively or recursively. It has two possible
     * values:
     * @returns The function `getNodeByKey` returns a node (`N`) if a node with the specified key is
     * found in the binary tree. If no node is found, it returns `undefined`.
     */
    getNodeByKey(key: BTNKey, iterationType?: IterationType): N | undefined;
    /**
     * The function `ensureNotKey` returns the node corresponding to the given key if it is a valid node
     * key, otherwise it returns the key itself.
     * @param {BTNKey | N | null | undefined} key - The `key` parameter can be of type `BTNKey`, `N`,
     * `null`, or `undefined`. It represents a key used to identify a node in a binary tree.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to be used when searching for a node by key. It has a default value of
     * `IterationType.ITERATIVE`.
     * @returns either the node corresponding to the given key if it is a valid node key, or the key
     * itself if it is not a valid node key.
     */
    ensureNotKey(key: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    get<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): V | undefined;
    get<C extends BTNCallback<N, N>>(identifier: N | null | undefined, callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): V | undefined;
    get<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): V | undefined;
    /**
     * Clear the binary tree, removing all nodes.
     */
    clear(): void;
    /**
     * Check if the binary tree is empty.
     * @returns {boolean} - True if the binary tree is empty, false otherwise.
     */
    isEmpty(): boolean;
    /**
     * The function `getPathToRoot` returns an array of nodes starting from a given node and traversing
     * up to the root node, with the option to reverse the order of the nodes.
     * @param {N} beginRoot - The `beginRoot` parameter represents the starting node from which you want
     * to find the path to the root node.
     * @param [isReverse=true] - The `isReverse` parameter is a boolean flag that determines whether the
     * resulting path should be reversed or not. If `isReverse` is set to `true`, the path will be
     * reversed before returning it. If `isReverse` is set to `false` or not provided, the path will
     * @returns The function `getPathToRoot` returns an array of type `N[]`.
     */
    getPathToRoot(beginRoot: BTNKey | N | null | undefined, isReverse?: boolean): N[];
    /**
     * The function `getLeftMost` returns the leftmost node in a binary tree, either using recursive or
     * iterative traversal.
     * @param {BTNKey | N | null | undefined} beginRoot - The `beginRoot` parameter is the starting point
     * for finding the leftmost node in a binary tree. It can be either a node object (`N`), a key value
     * of a node (`BTNKey`), or `null` if the tree is empty.
     * @param iterationType - The `iterationType` parameter is used to determine the type of iteration to
     * be performed when finding the leftmost node in a binary tree. It can have two possible values:
     * @returns The function `getLeftMost` returns the leftmost node (`N`) in a binary tree. If there is
     * no leftmost node, it returns `null`.
     */
    getLeftMost(beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    /**
     * The function `getRightMost` returns the rightmost node in a binary tree, either recursively or
     * iteratively.
     * @param {N | null | undefined} beginRoot - The `beginRoot` parameter is the starting node from which we want to
     * find the rightmost node. It is of type `N | null | undefined`, which means it can either be a node of type `N`
     * or `null`. If it is `null`, it means there is no starting node
     * @param iterationType - The `iterationType` parameter is used to determine the type of iteration to
     * be performed when finding the rightmost node in a binary tree. It can have two possible values:
     * @returns The function `getRightMost` returns the rightmost node (`N`) in a binary tree. If the
     * `beginRoot` parameter is `null`, it returns `null`.
     */
    getRightMost(beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType): N | null | undefined;
    /**
     * The function `isSubtreeBST` checks if a given binary tree is a valid binary search tree.
     * @param {N} beginRoot - The `beginRoot` parameter is the root node of the binary tree that you want
     * to check if it is a binary search tree (BST) subtree.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when checking if a subtree is a binary search tree (BST). It can have two
     * possible values:
     * @returns The function `isSubtreeBST` returns a boolean value.
     */
    isSubtreeBST(beginRoot: BTNKey | N | null | undefined, iterationType?: IterationType): boolean;
    /**
     * The function checks if a binary tree is a binary search tree.
     * @param iterationType - The parameter "iterationType" is used to specify the type of iteration to
     * be used when checking if the binary tree is a binary search tree (BST). It is an optional
     * parameter with a default value of "this.iterationType". The value of "this.iterationType" is not
     * provided in
     * @returns a boolean value.
     */
    isBST(iterationType?: IterationType): boolean;
    subTreeTraverse<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: false): ReturnType<C>[];
    subTreeTraverse<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: undefined): ReturnType<C>[];
    subTreeTraverse<C extends BTNCallback<N | null | undefined>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: true): ReturnType<C>[];
    /**
     * The function checks if a given node is a real node by verifying if it is an instance of
     * BinaryTreeNode and its key is not NaN.
     * @param {any} node - The parameter `node` is of type `any`, which means it can be any data type.
     * @returns a boolean value.
     */
    isRealNode(node: any): node is N;
    /**
     * The function checks if a given node is a BinaryTreeNode instance and has a key value of NaN.
     * @param {any} node - The parameter `node` is of type `any`, which means it can be any data type.
     * @returns a boolean value.
     */
    isNIL(node: any): boolean;
    /**
     * The function checks if a given node is a real node or null.
     * @param {any} node - The parameter `node` is of type `any`, which means it can be any data type.
     * @returns a boolean value.
     */
    isNodeOrNull(node: any): node is (N | null);
    /**
     * The function "isNodeKey" checks if a potential key is a number.
     * @param {any} potentialKey - The potentialKey parameter is of type any, which means it can be any
     * data type.
     * @returns a boolean value indicating whether the potentialKey is of type number or not.
     */
    isNodeKey(potentialKey: any): potentialKey is number;
    dfs<C extends BTNCallback<N>>(callback?: C, pattern?: DFSOrderPattern, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: false): ReturnType<C>[];
    dfs<C extends BTNCallback<N>>(callback?: C, pattern?: DFSOrderPattern, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: undefined): ReturnType<C>[];
    dfs<C extends BTNCallback<N | null | undefined>>(callback?: C, pattern?: DFSOrderPattern, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: true): ReturnType<C>[];
    bfs<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: false): ReturnType<C>[];
    bfs<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: undefined): ReturnType<C>[];
    bfs<C extends BTNCallback<N | null | undefined>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: true): ReturnType<C>[];
    listLevels<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: false): ReturnType<C>[][];
    listLevels<C extends BTNCallback<N>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: undefined): ReturnType<C>[][];
    listLevels<C extends BTNCallback<N | null | undefined>>(callback?: C, beginRoot?: BTNKey | N | null | undefined, iterationType?: IterationType, includeNull?: true): ReturnType<C>[][];
    getPredecessor(node: N): N;
    /**
     * The function `getSuccessor` returns the next node in a binary tree given a node `x`, or `null` if
     * `x` is the last node.
     * @param {N} x - N - a node in a binary tree
     * @returns The function `getSuccessor` returns a value of type `N` (the successor node), or `null`
     * if there is no successor, or `undefined` if the input `x` is `undefined`.
     */
    getSuccessor(x?: BTNKey | N | null): N | null | undefined;
    /**
     * The `morris` function performs a depth-first traversal of a binary tree using the Morris traversal
     * algorithm and returns an array of values obtained by applying a callback function to each node.
     * @param callback - The `callback` parameter is a function that will be called on each node in the
     * tree. It takes a node of type `N` as input and returns a value of type `ReturnType<BTNCallback<N>>`. The
     * default value for this parameter is `this._defaultOneParamCallback`.
     * @param {DFSOrderPattern} [pattern=in] - The `pattern` parameter in the `morris` function
     * determines the order in which the nodes of a binary tree are traversed. It can have one of the
     * following values:
     * @param {N | null | undefined} beginRoot - The `beginRoot` parameter is the starting node for the Morris
     * traversal. It specifies the root node of the tree from which the traversal should begin. If
     * `beginRoot` is `null`, an empty array will be returned.
     * @returns The `morris` function returns an array of `ReturnType<BTNCallback<N>>` values.
     */
    morris<C extends BTNCallback<N>>(callback?: C, pattern?: DFSOrderPattern, beginRoot?: BTNKey | N | null | undefined): ReturnType<C>[];
    /**
     * The above function is an iterator for a binary tree that can be used to traverse the tree in
     * either an iterative or recursive manner.
     * @param node - The `node` parameter represents the current node in the binary tree from which the
     * iteration starts. It is an optional parameter with a default value of `this.root`, which means
     * that if no node is provided, the iteration will start from the root of the binary tree.
     * @returns The `*[Symbol.iterator]` method returns a generator object that yields the keys of the
     * binary tree nodes in a specific order.
     */
    [Symbol.iterator](node?: N | null | undefined): Generator<BTNKey, void, undefined>;
    protected _defaultOneParamCallback: (node: N) => number;
    /**
     * Swap the data of two nodes in the binary tree.
     * @param {N} srcNode - The source node to swap.
     * @param {N} destNode - The destination node to swap.
     * @returns {N} - The destination node after the swap.
     */
    protected _swap(srcNode: BTNKey | N | null | undefined, destNode: BTNKey | N | null | undefined): N | undefined;
    /**
     * The function `_addTo` adds a new node to a binary tree if there is an available position.
     * @param {N | null | undefined} newNode - The `newNode` parameter represents the node that you want to add to
     * the binary tree. It can be either a node object or `null`.
     * @param {N} parent - The `parent` parameter represents the parent node to which the new node will
     * be added as a child.
     * @returns either the left or right child node of the parent node, depending on which child is
     * available for adding the new node. If a new node is added, the function also updates the size of
     * the binary tree. If neither the left nor right child is available, the function returns undefined.
     * If the parent node is null, the function also returns undefined.
     */
    protected _addTo(newNode: N | null | undefined, parent: BTNKey | N | null | undefined): N | null | undefined;
    /**
     * The function sets the root property of an object to a given value, and if the value is not null,
     * it also sets the parent property of the value to undefined.
     * @param {N | null | undefined} v - The parameter `v` is of type `N | null | undefined`, which means it can either be of
     * type `N` or `null`.
     */
    protected _setRoot(v: N | null | undefined): void;
    /**
     * The `print` function is used to display a binary tree structure in a visually appealing way.
     * @param {N | null | undefined} root - The `root` parameter in the `print` function represents the
     * root node of a binary tree. It can have one of the following types: `BTNKey`, `N`, `null`, or
     * `undefined`. The default value is `this.root`, which suggests that `this.root` is the
     */
    print(beginRoot?: BTNKey | N | null | undefined): void;
}
