/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BSTComparator, BSTNodeNested, BSTOptions, BTNCallback, BTNKey } from '../../types';
import { CP, IterationType } from '../../types';
import { BinaryTree, BinaryTreeNode } from './binary-tree';
import { IBinaryTree } from '../../interfaces';
export declare class BSTNode<V = any, N extends BSTNode<V, N> = BSTNodeNested<V>> extends BinaryTreeNode<V, N> {
    parent?: N;
    constructor(key: BTNKey, value?: V);
    protected _left?: N;
    /**
     * Get the left child node.
     */
    get left(): N | undefined;
    /**
     * Set the left child node.
     * @param {N | undefined} v - The left child node.
     */
    set left(v: N | undefined);
    protected _right?: N;
    /**
     * Get the right child node.
     */
    get right(): N | undefined;
    /**
     * Set the right child node.
     * @param {N | undefined} v - The right child node.
     */
    set right(v: N | undefined);
}
export declare class BST<V = any, N extends BSTNode<V, N> = BSTNode<V, BSTNodeNested<V>>> extends BinaryTree<V, N> implements IBinaryTree<V, N> {
    /**
     * The constructor function initializes a binary search tree object with an optional comparator
     * function.
     * @param {BSTOptions} [options] - An optional object that contains configuration options for the
     * binary search tree.
     */
    constructor(options?: BSTOptions);
    protected _root?: N;
    /**
     * Get the root node of the binary tree.
     */
    get root(): N | undefined;
    /**
     * The function creates a new binary search tree node with the given key and value.
     * @param {BTNKey} key - The key parameter is the key value that will be associated with
     * the new node. It is used to determine the position of the node in the binary search tree.
     * @param [value] - The parameter `value` is an optional value that can be assigned to the node. It
     * represents the value associated with the node in a binary search tree.
     * @returns a new instance of the BSTNode class with the specified key and value.
     */
    createNode(key: BTNKey, value?: V): N;
    /**
     * The `add` function in a binary search tree class inserts a new node with a given key and value
     * into the tree.
     * @param {BTNKey | N | undefined} keyOrNode - The `keyOrNode` parameter can be either a
     * `BTNKey` (which can be a number or a string), a `BSTNode` object, or `undefined`.
     * @param [value] - The `value` parameter is the value to be assigned to the new node being added to the
     * binary search tree.
     * @returns the inserted node (N) if it was successfully added to the binary search tree. If the node
     * was not added or if the parameters were invalid, it returns undefined or undefined.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    /**
     * The `addMany` function is used to efficiently add multiple nodes to a binary search tree while
     * maintaining balance.
     * @param {[BTNKey | N, V][]} keysOrNodes - The `arr` parameter in the `addMany` function
     * represents an array of keys or nodes that need to be added to the binary search tree. It can be an
     * array of `BTNKey` or `N` (which represents the node type in the binary search tree) or
     * `undefined
     * @param {V[]} data - The values of tree nodes
     * @param {boolean} isBalanceAdd - If true the nodes will be balance inserted in binary search method.
     * @param iterationType - The `iterationType` parameter determines the type of iteration to be used.
     * It can have two possible values:
     * @returns The `addMany` function returns an array of `N`, `undefined`, or `undefined` values.
     */
    addMany(keysOrNodes: (BTNKey | N | undefined)[], data?: (V | undefined)[], isBalanceAdd?: boolean, iterationType?: IterationType): (N | undefined)[];
    /**
     * The function `lastKey` returns the key of the rightmost node if the comparison result is less
     * than, the key of the leftmost node if the comparison result is greater than, and the key of the
     * rightmost node otherwise.
     * @param {N | undefined} beginRoot - The `beginRoot` parameter is the starting point for finding the last
     * key in a binary tree. It represents the root node of the subtree from which the search for the
     * last key should begin. If no specific `beginRoot` is provided, the search will start from the root
     * of the entire binary
     * @param iterationType - The `iterationType` parameter is used to specify the type of iteration to
     * be performed when finding the last key. It determines whether the iteration should be performed in
     * pre-order, in-order, or post-order.
     * @returns the key of the rightmost node in the binary tree if the comparison result is less than,
     * the key of the leftmost node if the comparison result is greater than, and the key of the
     * rightmost node otherwise. If no node is found, it returns 0.
     */
    lastKey(beginRoot?: BTNKey | N | undefined, iterationType?: IterationType): BTNKey;
    /**
     * The function `getNodeByKey` searches for a node in a binary tree based on a given key, using
     * either recursive or iterative methods.
     * @param {BTNKey} key - The `key` parameter is the key value that we are searching for in the tree.
     * It is used to find the node with the matching key value.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when searching for a node in the binary tree. It can have two possible
     * values:
     * @returns The function `getNodeByKey` returns a node (`N`) if a node with the specified key is
     * found in the binary tree. If no node is found, it returns `undefined`.
     */
    getNodeByKey(key: BTNKey, iterationType?: IterationType): N | undefined;
    /**
     * The function `ensureNotKey` returns the node corresponding to the given key if it is a node key,
     * otherwise it returns the key itself.
     * @param {BTNKey | N | undefined} key - The `key` parameter can be of type `BTNKey`, `N`, or
     * `undefined`.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to be performed. It has a default value of `IterationType.ITERATIVE`.
     * @returns either a node object (N) or undefined.
     */
    ensureNotKey(key: BTNKey | N | undefined, iterationType?: IterationType): N | undefined;
    /**
     * The function `getNodes` retrieves nodes from a binary tree based on a given node property or key,
     * using either recursive or iterative traversal.
     * @param {ReturnType<C> | N} identifier - The `nodeProperty` parameter represents the property
     * of the binary tree node that you want to search for. It can be either a `BTNKey` or a
     * generic type `N`.
     * @param callback - The `callback` parameter is a function that takes a node as input and returns a
     * value. This value is compared with the `nodeProperty` parameter to determine if the node should be
     * included in the result. The default value for `callback` is `this._defaultOneParamCallback`, which is
     * a
     * @param [onlyOne=false] - A boolean value indicating whether to stop the traversal after finding
     * the first node that matches the nodeProperty. If set to true, the function will return an array
     * containing only that node. If set to false (default), the function will continue the traversal and
     * return an array containing all nodes that match the node
     * @param {N | undefined} beginRoot - The `beginRoot` parameter is the starting node for the traversal. It
     * specifies the root node of the binary tree from which the traversal should begin. If `beginRoot`
     * is `undefined`, an empty array will be returned.
     * @param iterationType - The `iterationType` parameter determines the type of iteration used to
     * traverse the binary tree. It can have one of the following values:
     * @returns an array of nodes (N[]).
     */
    getNodes<C extends BTNCallback<N>>(identifier: ReturnType<C> | undefined, callback?: C, onlyOne?: boolean, beginRoot?: BTNKey | N | undefined, iterationType?: IterationType): N[];
    /**
     * The `lesserOrGreaterTraverse` function traverses a binary tree and applies a callback function to
     * nodes that have a key value lesser or greater than a target key value.
     * @param callback - The `callback` parameter is a function that will be called for each node that
     * meets the condition specified by the `lesserOrGreater` parameter. It takes a node as an argument
     * and returns a value.
     * @param {CP} lesserOrGreater - The `lesserOrGreater` parameter is used to determine whether to
     * traverse nodes that are lesser than, greater than, or equal to the `targetNode`. It can take one
     * of the following values:
     * @param {BTNKey | N | undefined} targetNode - The `targetNode` parameter in the
     * `lesserOrGreaterTraverse` function is used to specify the node from which the traversal should
     * start. It can be either a reference to a specific node (`N`), the key of a node
     * (`BTNKey`), or `undefined` to
     * @param iterationType - The `iterationType` parameter determines whether the traversal should be
     * done recursively or iteratively. It can have two possible values:
     * @returns The function `lesserOrGreaterTraverse` returns an array of `ReturnType<BTNCallback<N>>`.
     */
    lesserOrGreaterTraverse<C extends BTNCallback<N>>(callback?: C, lesserOrGreater?: CP, targetNode?: BTNKey | N | undefined, iterationType?: IterationType): ReturnType<C>[];
    /**
     * Balancing Adjustment:
     * Perfectly Balanced Binary Tree: Since the balance of a perfectly balanced binary tree is already fixed, no additional balancing adjustment is needed. Any insertion or deletion operation will disrupt the perfect balance, often requiring a complete reconstruction of the tree.
     * AVL Tree: After insertion or deletion operations, an AVL tree performs rotation adjustments based on the balance factor of nodes to restore the tree's balance. These rotations can be left rotations, right rotations, left-right rotations, or right-left rotations, performed as needed.
     *
     * Use Cases and Efficiency:
     * Perfectly Balanced Binary Tree: Perfectly balanced binary trees are typically used in specific scenarios such as complete binary heaps in heap sort or certain types of Huffman trees. However, they are not suitable for dynamic operations requiring frequent insertions and deletions, as these operations often necessitate full tree reconstruction.
     * AVL Tree: AVL trees are well-suited for scenarios involving frequent searching, insertion, and deletion operations. Through rotation adjustments, AVL trees maintain their balance, ensuring average and worst-case time complexity of O(log n).
     */
    /**
     * The `perfectlyBalance` function balances a binary search tree by adding nodes in a way that
     * ensures the tree is perfectly balanced.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building a balanced binary search tree. It can have two possible
     * values:
     * @returns The function `perfectlyBalance` returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * The function checks if a binary tree is AVL balanced using either recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is used to determine the method of iteration
     * to check if the AVL tree is balanced. It can have two possible values:
     * @returns a boolean value.
     */
    isAVLBalanced(iterationType?: IterationType): boolean;
    protected _comparator: BSTComparator;
    protected _setRoot(v: N | undefined): void;
    /**
     * The function compares two values using a comparator function and returns whether the first value
     * is greater than, less than, or equal to the second value.
     * @param {BTNKey} a - The parameter "a" is of type BTNKey.
     * @param {BTNKey} b - The parameter "b" in the above code represents a BTNKey.
     * @returns a value of type CP (ComparisonResult). The possible return values are CP.gt (greater
     * than), CP.lt (less than), or CP.eq (equal).
     */
    protected _compare(a: BTNKey, b: BTNKey): CP;
}
