/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { BiTreeDeleteResult, BTNCallback, BTNKey, IterationType, RBTNColor, RedBlackTreeNodeNested, RBTreeOptions } from '../../types';
import { BST, BSTNode } from "./bst";
import { IBinaryTree } from "../../interfaces";
export declare class RedBlackTreeNode<V = any, N extends RedBlackTreeNode<V, N> = RedBlackTreeNodeNested<V>> extends BSTNode<V, N> {
    color: RBTNColor;
    constructor(key: BTNKey, value?: V, color?: RBTNColor);
}
/**
 * 1. Each node is either red or black.
 * 2. The root node is always black.
 * 3. Leaf nodes are typically NIL nodes and are considered black.
 * 4. Red nodes must have black children.
 * 5. Black balance: Every path from any node to each of its leaf nodes contains the same number of black nodes.
 */
export declare class RedBlackTree<V = any, N extends RedBlackTreeNode<V, N> = RedBlackTreeNode<V, RedBlackTreeNodeNested<V>>> extends BST<V, N> implements IBinaryTree<V, N> {
    constructor(options?: RBTreeOptions);
    protected _root: N;
    get root(): N;
    protected _size: number;
    get size(): number;
    NIL: N;
    /**
     * The `add` function adds a new node to a Red-Black Tree data structure.
     * @param {BTNKey | N | null | undefined} keyOrNode - The `keyOrNode` parameter can be one of the
     * following types:
     * @param {V} [value] - The `value` parameter is an optional value that can be associated with the
     * key in the node being added to the Red-Black Tree.
     * @returns The method returns either a node (`N`) or `undefined`.
     */
    add(keyOrNode: BTNKey | N | null | undefined, value?: V): N | undefined;
    createNode(key: BTNKey, value?: V, color?: RBTNColor): N;
    /**
     * The `delete` function removes a node from a binary tree based on a given identifier and updates
     * the tree accordingly.
     * @param {ReturnType<C> | null | undefined} identifier - The `identifier` parameter is the value
     * that you want to use to identify the node that you want to delete from the binary tree. It can be
     * of any type that is returned by the callback function `C`. It can also be `null` or `undefined` if
     * you don't want to
     * @param {C} callback - The `callback` parameter is a function that takes a node of type `N` and
     * returns a value of type `ReturnType<C>`. It is used to determine if a node should be deleted based
     * on its identifier. The `callback` function is optional and defaults to `this._defaultOneParam
     * @returns an array of `BiTreeDeleteResult<N>`.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C> | null | undefined, callback?: C): BiTreeDeleteResult<N>[];
    isRealNode(node: N | undefined): node is N;
    getNode<C extends BTNCallback<N, BTNKey>>(identifier: BTNKey, callback?: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    getNode<C extends BTNCallback<N, N>>(identifier: N | undefined, callback?: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    getNode<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback: C, beginRoot?: N | undefined, iterationType?: IterationType): N | undefined;
    /**
     * The function returns the successor of a given node in a red-black tree.
     * @param {RedBlackTreeNode} x - RedBlackTreeNode - The node for which we want to find the successor.
     * @returns the successor of the given RedBlackTreeNode.
     */
    getSuccessor(x: N): N | undefined;
    /**
     * The function returns the predecessor of a given node in a red-black tree.
     * @param {RedBlackTreeNode} x - The parameter `x` is of type `RedBlackTreeNode`, which represents a node in a
     * Red-Black Tree.
     * @returns the predecessor of the given RedBlackTreeNode 'x'.
     */
    getPredecessor(x: N): N;
    clear(): void;
    protected _setRoot(v: N): void;
    /**
     * The function performs a left rotation on a red-black tree node.
     * @param {RedBlackTreeNode} x - The parameter `x` is a RedBlackTreeNode object.
     */
    protected _leftRotate(x: N): void;
    /**
     * The function performs a right rotation on a red-black tree node.
     * @param {RedBlackTreeNode} x - x is a RedBlackTreeNode, which represents the node that needs to be right
     * rotated.
     */
    protected _rightRotate(x: N): void;
    /**
     * The _fixDelete function is used to rebalance the Red-Black Tree after a node deletion.
     * @param {RedBlackTreeNode} x - The parameter `x` is of type `RedBlackTreeNode`, which represents a node in a
     * red-black tree.
     */
    protected _fixDelete(x: N): void;
    /**
     * The function `_rbTransplant` replaces one node in a red-black tree with another node.
     * @param {RedBlackTreeNode} u - The parameter "u" represents a RedBlackTreeNode object.
     * @param {RedBlackTreeNode} v - The parameter "v" is a RedBlackTreeNode object.
     */
    protected _rbTransplant(u: N, v: N): void;
    /**
     * The `_fixInsert` function is used to fix the red-black tree after an insertion operation.
     * @param {RedBlackTreeNode} k - The parameter `k` is a RedBlackTreeNode object, which represents a node in a
     * red-black tree.
     */
    protected _fixInsert(k: N): void;
}
